import json
import os
import pdb

from collections import OrderedDict

class ReadOnlyDict(OrderedDict):
    def __missing__(self, key):
        raise KeyError(f"'{key}' not found in this object. List of keys = {self.keys()}.")

    def __getattr__(self, item): 
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item) # to allow deepcopy (from s/o)
        return self[item]

    def __setattr__(self, key, value):
        #if key not in self and "Config" not in key and "List" not in key: # these are lame; find way in schema to initialize complex types {}, [], or null
        if key not in self: # these are lame; find way in schema to initialize complex types {}, [], or null
            self.__missing__(key) # this should not be necessary

        if "schema" in self:
            if "type" in self["schema"][key]:
                if self["schema"][key]["type"] in [ "integer", "float" ]:
                    if type(value) is str:
                        raise ValueError( "{} is string when needs to be {} for parameter {}.".format( value, self["schema"][key]["type"], key ) )
                    elif value < self["schema"][key]["min"]:
                        #print( "{} is below minimum {} for parameter {}.".format( value, self["schema"][key]["min"], key ) )
                        raise ValueError( "{} is below minimum {} for parameter {}.".format( value, self["schema"][key]["min"], key ) )
                    elif value > self["schema"][key]["max"]:
                        #print( "{} is above maximum {} for parameter {}.".format( value, self["schema"][key]["max"], key ) )
                        raise ValueError( "{} is above maximum {} for parameter {}.".format( value, self["schema"][key]["max"], key ) )

                elif self["schema"][key]["type"] == "enum":
                    if value not in self["schema"][key]["enum"]:
                        #print( "{} is not list of possible values {} for parameter {}.".format( value, self["schema"][key]["enum"] ) )
                        raise ValueError( "{} is not list of possible values {} for parameter {}.".format( value, self["schema"][key]["enum"], key ) )
                elif self["schema"][key]["type"] == "bool":
                    if not any( [ value == False, value == True, value == 1, value == 0 ] ):
                        #print( "Value needs to be a bool for parameter {key}." )
                        raise ValueError( f"value needs to be a bool for parameter {key}." )
                    if value == False:
                        value = 0
                    elif value == True:
                        value = 1
                elif "Vector" in self["schema"][key]["type"]:
                    if type(value) is not list:
                        raise ValueError( f"value needs to be a list for parameter {key}." )
            else:
                print( "This better be a blob of tabular data. Allow a hack here for Malaria_Drug_Parans, Vector_Species_Params, etc. until we fix the schema." )
                # allow the assignment of a python dict here.
                pass

        # hold on: if param is dependent on a param, set that param to the value. but needs to be recursive.
        if "schema" not in self:
            print( f"DEBUG: No schema in node for param {key}." )
        elif "depends-on" in self["schema"][key]:
            # set this value BUT NOT IF key == Simulation_Type!
            for k,v in dict(self["schema"][key]["depends-on"]).items():
                if k == "Simulation_Type":
                    if k not in self.keys():
                        continue # not supported yet for campaigns; need to provide campaign blobs access to config...
                    elif self["Simulation_Type"] in [x.strip() for x in v.split(',')]:
                        pass
                        #print( "VERBOSE: Simulation_Type already set to valid value for this param." )
                    else:
                        print( f"ERROR: Simulation_Type needs to be one of {v} for you to be able to set {key} to {value} but it seems to be {self.Simulation_Type}." )
                        raise ValueError( f"ERROR: Simulation_Type needs to be one of {v} for you to be able to set {key} to {value} but it seems to be {self.Simulation_Type}." )
                else:
                    if type(v) is str and len(v.split(','))>1:
                        v = v.split(',')[0] # pretty arbitrary but least arbitrary of options it seems
                    #print( f"IMPLICITLY setting {k} to {v}." )
                    self[k] = v
                    if "implicits" not in self: # This should NOT be needed
                        self["implicits"] = [] 
                    self["implicits"].append( key )

        self[key] = value
        if "explicits" not in self: # This should NOT be needed
            self["explicits"] = [] 

        self["explicits"].append( key )
        #print( "VERBOSE: Explicits = " + str( self["explicits"] ) )
        return

    def set_schema(self,schema):
        self["schema"] = schema

    def to_file(self, config_name="config.json"):
        """
        Write 'clean' config file out to disk as json. 
        Param: config_name (defaults to 'config.json')
        """
        config = self.finalize()
        with open( config_name, "w" ) as config_file:
            json.dump( config, config_file, indent=4, sort_keys=True )

    def finalize(self):
        # Remove all params that are disabled by depends-on param being off
        nuke_list = []
        for key, v in self.items():
            if key in [ "schema", "explicits", "implicits" ]:
                continue
            elif key not in self["schema"]:
                print( f"WARNING: During schema-based param purge, {key} not in schema." )
            elif "depends-on" in self["schema"][key]:
                for dep_k,dep_v in dict(self["schema"][key]["depends-on"]).items():
                    if type(dep_v) is str:
                        vals = [x.strip() for x in str(dep_v).split(',')] 
                        if dep_k in self and self[dep_k] not in vals:
                            #print( f"VERBOSE: Purge {key} from raw config because {dep_k} value of {self[dep_k]} not in {vals}." )
                            nuke_list.append( key )
                    else:
                        if self[dep_k] != dep_v:
                            #print( f"VERBOSE: Purge {key} from raw config because {dep_k} value of {self[dep_k]} not in {dep_v}." )
                            nuke_list.append( key )
            elif v == "UNINITIALIZED STRING": # work around current schema string defaults
                self[key] = ""
        # Note that nuke_list is not a set and is typically full of duplicates. There are many ways to de-dupe.
        #print( str( nuke_list ) )
        for nuke_key in nuke_list:
            #print( f"VERBOSE: Purging param {nuke_key}." )
            if nuke_key in self.keys():
                if "explicits" in self and nuke_key in self['explicits']:
                    #print( f"You set param {nuke_key} but it was disabled and is not being used." )
                    raise ValueError( f"You set param {nuke_key} but it was disabled and is not being used." )
                self.pop( nuke_key )
        if "implicits" in self:
            #for key in self["implicits"]:
                #print( "VERBOSE: Parameter {} set implicitly.".format( key ) )
            self.pop( "implicits" )
        if "explicits" in self:
            self.pop( "explicits" )
        try:
            #print( "VERBOSE: Purging schema, etc." )
            self.pop( "schema" )
        except Exception as ex:
            #pdb.set_trace()
            print( "ERROR: " + str( ex ) )
            raise ValueError( f"ERROR: Something bad happened during finalize: {ex}." )
        return self

def _get_default_campaign_event():
    """
    This function was part of original version but should no longer be used.
    Use get_class_with_defaults now with classname="CampaignEvent".
    """
    event = {}
    event = json.loads(
        """
        {
            "Event_Coordinator_Config": {},
            "Nodeset_Config": { "class": "NodeSetAll" },
            "Start_Day": -1,
            "class": "CampaignEvent"
            }
        """,
        object_hook=ReadOnlyDict
    )
    return event


def get_default_for_complex_type( schema, idmtype ):
    # get type info from schema
    type_json = schema[idmtype]
    if type(type_json) is dict:
        # I think we can/should do something recursive here.
        return dict()
    elif type(type_json) is list:
        return list()
    else:
        print( f"WARNING: Failed to figure out how to make outer default (dict or list) for idmtype {idmtype}." )
        return dict()


def get_class_with_defaults( classname, schema_path=None ):
    """
    Very handy little class -- that needs a lot more error handling and robustification --
    that gets you the default config for an event coordinator or intervention from the schema.
    """
    #print( f"VERBOSE: get_class_with_defaults called with schema_path = {schema_path}." )
    if schema_path is None:
        schema_path = "schema.json" 
    if not os.path.exists( schema_path ):
        print( f"ERROR: No file found at {schema_path}. A valid schema path needs to exist at the path specified." )
        raise ValueError( f"ERROR: No file found at {schema_path}. A valid schema path needs to exist at the path specified." )
        return None

    #print( f"DEBUG: get_class_with_defaults called with {classname} and {schema_path}." )
    ret_json = {}
    schema = json.loads( open( schema_path ).read() )

    schema_blob = None
    if "campaignevent" in classname.lower():
        if classname in schema["idmTypes"]["idmAbstractType:CampaignEvent"].keys():
            schema_blob = schema["idmTypes"]["idmAbstractType:CampaignEvent"][classname]
            ret_json["class"] = schema_blob["class"]
            for ce_key in schema_blob.keys():
                if ce_key == "class":
                    continue
                try:
                    if "default" in schema_blob[ce_key] and schema_blob[ce_key]["default"] != "null":
                        ret_json[ce_key] = schema_blob[ce_key]["default"]
                    elif ce_key == "Nodeset_Config": # this doesn't look a real pattern
                        ret_json[ce_key] = get_class_with_defaults( "NodeSetAll", schema_path )
                    elif ce_key != "class":
                        ret_json[ce_key] = {}
                except Exception as ex:
                    #pdb.set_trace()
                    print( "ERROR: " + str( ex ) )
                    raise ValueError( f"ERROR: " + str( ex ) )
    elif "coordinator" in classname.lower():
        for ec_name in schema["idmTypes"]["idmAbstractType:EventCoordinator"].keys():
            if ec_name == classname or classname.replace( "EventCoordinator", "" ) in ec_name:
                #print( schema["idmTypes"]["idmAbstractType:EventCoordinator"][ec_name] )
                schema_blob = schema["idmTypes"]["idmAbstractType:EventCoordinator"][ec_name]
                ret_json["class"] = schema_blob["class"]
                for ec_key in schema_blob.keys():
                    if ec_key == "class":
                        continue
                    try:
                        if "default" in schema_blob[ec_key]:
                            ret_json[ec_key] = schema_blob[ec_key]["default"]
                            # have to hack a couple of things for now, TBD: fix in schema
                            if ec_key == "Property_Restrictions_Within_Node" or ec_key == "Node_Property_Restrictions":
                                ret_json[ec_key] = []
                        elif "Config" in ec_key:
                            ret_json[ec_key] = {}
                    except Exception as ex:
                        #pdb.set_trace()
                        print( "ERROR: " + str( ex ) )
                        raise ValueError( f"ERROR: " + str( ex ) )
                break # once we find it, stop
    elif "waning" in classname.lower():
        #for wan_type in schema["idmTypes"]["idmType:WaningEffect"].keys():
        if classname in schema["idmTypes"]["idmType:WaningEffect"].keys():
            schema_blob = schema["idmTypes"]["idmType:WaningEffect"][classname]
            ret_json["class"] = schema_blob["class"]
            for wan_key in schema_blob.keys():
                if wan_key == "class":
                    continue
                try:
                    if "default" in schema_blob[wan_key]:
                        ret_json[wan_key] = schema_blob[wan_key]["default"]
                    elif "idmType:" in schema_blob[wan_key]["type"]:
                        # this probably needs a fancy solution, but in the meantime...
                        idmtype = schema_blob[wan_key]["type"].split( ":" )[1]
                        if "Map" in idmtype:
                            ret_json[wan_key] = {}
                        elif "List" in idmtype:
                            ret_json[wan_key] = []
                        else:
                            print( f"Don't know how to make default for type {idmtype}." )
                except Exception as ex:
                    #pdb.set_trace()
                    print( "ERROR: " + str( ex ) )
                    raise ValueError( f"ERROR: " + str( ex ) )
    elif "nodeset" in classname.lower():
        if classname in schema["idmTypes"]["idmAbstractType:NodeSet"].keys():
            schema_blob = schema["idmTypes"]["idmAbstractType:NodeSet"][classname]
            ret_json["class"] = schema_blob["class"]
            for ns_key in schema_blob.keys():
                if ns_key == "class":
                    continue
                try:
                    if "default" in schema_blob[ns_key]:
                        ret_json[ns_key] = schema_blob[ns_key]["default"]
                        #print( f"Setting {ns_key} to {ret_json[ns_key]}." )
                    elif "type" in schema_blob[ns_key]:
                        if schema_blob[ns_key]["type"] == "idmType:NodeListConfig": # hack for now, might be schema bug
                            ret_json[ns_key] = []
                            #print( f"Setting {ns_key} to {ret_json[ns_key]}." )
                        elif "Vector" in schema_blob[ns_key]["type"]:
                            ret_json[ns_key] = []
                            #print( f"Setting {ns_key} to {ret_json[ns_key]}." )
                        else:
                            print( f"'type' not found in schema_blob[{ns_key}]." )
                    else:
                        print( f"WARNING: Not setting default for NodeSet key {ns_key}." )
                except Exception as ex:
                    print( f"ERROR: Exception caught while processing {ns_key} in NodeSet family." )
                    print( "ERROR: " + str( ex ) )
                    raise ValueError( f"ERROR: Exception caught while processing {ns_key} in NodeSet family." )
    elif "report" in classname.lower():
        if classname in schema["idmTypes"]["idmType:IReport"].keys():
            schema_blob = schema["idmTypes"]["idmType:IReport"][classname]
            ret_json["class"] = schema_blob["class"]
            for ce_key in schema_blob.keys():
                if ce_key == "class":
                    continue
                try:
                    if "default" in schema_blob[ce_key] and schema_blob[ce_key]["default"] != "null":
                        ret_json[ce_key] = schema_blob[ce_key]["default"]
                    elif ce_key == "Nodeset_Config": # this doesn't look a real pattern
                        ret_json[ce_key] = get_class_with_defaults( "NodeSetAll", schema_path )
                    elif ce_key != "class":
                        ret_json[ce_key] = {}
                except Exception as ex:
                    print( "ERROR: " + str( ex ) )
                    raise ValueError( f"ERROR: " + str( ex ) )
    else:
        #print( schema["idmTypes"]["idmAbstractType:Intervention"].keys() )
        for iv_type in schema["idmTypes"]["idmAbstractType:Intervention"].keys():
            if classname in schema["idmTypes"]["idmAbstractType:Intervention"][iv_type].keys():
                #print( schema["idmTypes"]["idmAbstractType:Intervention"][iv_type][classname] )
                schema_blob = schema["idmTypes"]["idmAbstractType:Intervention"][iv_type][classname]
                ret_json["class"] = schema_blob["class"]
                for iv_key in schema_blob.keys():
                    if any( [ iv_key == "class", iv_key == "iv_type" ] ):
                        continue
                    try:
                        if "default" in schema_blob[iv_key]:
                            ret_json[iv_key] = schema_blob[iv_key]["default"]
                            if iv_key == "Property_Restrictions_Within_Node" or iv_key == "Node_Property_Restrictions":
                                ret_json[iv_key] = []
                        # Going to allow default dict and lists based on type or even name?
                        elif "Config" in iv_key:
                            ret_json[iv_key] = {}
                        elif "type" in schema_blob[iv_key]:
                            idmtype = schema_blob[iv_key]["type"]
                            if "Vector" in idmtype:
                                ret_json[iv_key] = []
                            elif "String" in idmtype:
                                ret_json[iv_key] = ""
                            elif "idmType:" in schema_blob[iv_key]["type"]:
                                ret_json[iv_key] = get_default_for_complex_type( schema["idmTypes"], idmtype )
                            elif "List" in iv_key:  # a bit lame: to handle Intervention_List which has bad schema bug
                                ret_json[iv_key] = []
                            else:
                                print( f"Don't know how to make default for type {idmtype}." )
                        else:
                            print( f"Don't know how to make default for key {iv_key}." )

                    except Exception as ex:
                        print( f"ERROR: Exception caught while processing {iv_key} in Intervention family." )
                        print( "ERROR: " + str( ex ) )
                        raise ValueError( f"ERROR: Exception caught while processing {iv_key} in Intervention family." )
                break
        if bool(ret_json) == False:
            raise ValueError( f"Failed to find intervention {classname} in schema." )

    ret_this = ReadOnlyDict( ret_json )
    #ret_this.schema = schema_blob 
    ret_this.set_schema( schema_blob )
    return ret_this
